import { Result, getLinesAround } from "./util";
import chalk from "chalk";

export interface ParseInput {
  source: string;
  position: number;
}

export interface ParseSuccess<T> {
  output: T;
  start: number;
  end: number;
}

export interface ParseError {
  expected?: string;
  start: number;
}

export interface Parser<Out> {
  (_: ParseInput): Result<ParseSuccess<Out>, ParseError>;
}

export const exact: (_: string) => Parser<string> = (needle: string) => (input: ParseInput) => {
  const start = input.position;
  const end = start + needle.length;
  const sliced = input.source.slice(start, end);
  return sliced === needle
    ? Result.ok({ output: sliced, start, end })
    : Result.err({ expected: `The string ${JSON.stringify(needle)}`, start });
};

export const regex: (_: RegExp) => Parser<string> = (re: RegExp) => (input: ParseInput) => {
  const start = input.position;
  const reY = new RegExp(re, [...new Set([...re.flags.split(""), "y", "m"])].join(""));

  reY.lastIndex = start;
  const result = input.source.match(reY);
  if (!result) return Result.err({ expected: `String that matches ${reY}`, start });
  return Result.ok({ output: result[0], start, end: start + result[0].length });
};

export const optWhitespace = regex(/\s*/m);
export const whitespace = regex(/\s+/m);

export const surroundWs = <T>(parser: Parser<T>) =>
  first(second(optWhitespace, parser), optWhitespace);

export const map =
  <T, U>(parser: Parser<T>, func: (_: T) => U) =>
  (input: ParseInput) =>
    parser(input).map(({ output, ...result }) => ({ output: func(output), ...result }));

export const pair: <Out1, Out2>(p1: Parser<Out1>, p2: Parser<Out2>) => Parser<[Out1, Out2]> =
  (p1, p2) => (input: ParseInput) =>
    p1(input).andThen(({ output: out1, end }) =>
      p2({ ...input, position: end }).map(({ output: out2, end }) => ({
        output: [out1, out2],
        start: input.position,
        end: end,
      })),
    );

export const first: <Out1, Out2>(p1: Parser<Out1>, p2: Parser<Out2>) => Parser<Out1> = (p1, p2) =>
  map(pair(p1, p2), ([out1]) => out1);

export const firstWs: <Out1, Out2>(p1: Parser<Out1>, p2: Parser<Out2>) => Parser<Out1> = (p1, p2) =>
  first(first(p1, whitespace), p2);

export const second: <Out1, Out2>(p1: Parser<Out1>, p2: Parser<Out2>) => Parser<Out2> = (p1, p2) =>
  map(pair(p1, p2), ([_, out2]) => out2);

export const secondWs: <Out1, Out2>(p1: Parser<Out1>, p2: Parser<Out2>) => Parser<Out2> = (
  p1,
  p2,
) => second(second(p1, whitespace), p2);

export const sep: <T, D>(itemParser: Parser<T>, delimParser: Parser<D>) => Parser<T[]> =
  (itemParser, delimParser) => (input: ParseInput) => {
    const items = [];
    let position = input.position;
    let first = true;
    while (true) {
      if (first) {
        first = false;
      } else {
        const result = delimParser({ ...input, position });
        if (result.status === "err") break;
        const { end } = result.unwrap();
        position = end;
      }

      const result = itemParser({ ...input, position });
      if (result.status === "err") break;
      const { output, end } = result.unwrap();
      position = end;
      items.push(output);
    }
    return Result.ok({ output: items, start: input.position, end: position });
  };

export const repeatWs: <T>(parser: Parser<T>) => Parser<T[]> = (parser) => sep(parser, whitespace);

export const alt =
  <T>(...parsers: Parser<T>[]) =>
  (input: ParseInput) =>
    parsers.reduceRight(
      (prevValue: Result<ParseSuccess<T>, ParseError>, nextValue: Parser<T>) =>
        prevValue.orElse(() => nextValue(input)),
      Result.err({ start: input.position }),
    );

// Language-specific stuff

export const ident = regex(/[_a-zA-Z][_a-zA-Z0-9]+/);

export const extern = map(first(secondWs(exact("extern"), ident), exact(";")), (name) => ({
  kind: "externItem",
  name,
}));

export const item = alt(extern);

export const program = surroundWs(repeatWs(item));

export function parseProgram(input: string) {
  const result = program({ source: input, position: 0 });
  if (result.status === "err") {
    const { expected, start } = result.unwrapErr();
    const expectedMsg = expected ? `Expected: ${expected}` : "";
    printError(input, start, expectedMsg);
  }

  const { output, end } = result.unwrap();
  if (end !== input.length) {
    printError(input, end, "wtf is this bro");
  }

  return output;
}

function printError(input: string, at: number, msg: string) {
  const linesAround = getLinesAround(input, at);
  if (linesAround.status === "err") throw new Error(msg);

  const { lines, focusedLineIdx, startLineNum, col } = linesAround.unwrap();

  const maxLen = Math.max(
    startLineNum.toString().length,
    (startLineNum + lines.length).toString().length,
  );

  console.log(chalk.red("Error parsing program:"));

  lines.forEach((line, idx) => {
    const absLine = idx + startLineNum;
    const absLineStr = absLine.toString().padStart(maxLen, " ");

    console.log(chalk.blue(`${absLineStr} | `) + line);
    if (idx === focusedLineIdx) {
      const emptyLineStr = "".padStart(maxLen, " ");
      const errorPointer = new Array(col).fill(0).map((_) => " ") + "^";
      console.log(chalk.blue(`${emptyLineStr} | `) + chalk.red(errorPointer + " " + msg));
    }
  });

  console.log();

  throw new Error();
}
