export class Result<T, E> {
  private constructor(
    public status: "ok" | "err",
    private okValue: T | undefined,
    private errValue: E | undefined,
  ) {}

  toString(): string {
    switch (this.status) {
      case "ok":
        return `Ok(${this.okValue!.toString()})`;
      case "err":
        return `Err(${this.errValue!.toString()})`;
    }
  }

  static ok<T, E>(value: T): Result<T, E> {
    return new Result("ok", value, undefined!);
  }

  static err<T, E>(value: E): Result<T, E> {
    return new Result("err", undefined!, value);
  }

  unwrap(): T {
    switch (this.status) {
      case "ok":
        return this.okValue!;
      case "err":
        throw new Error("unwrap");
    }
  }

  unwrapErr(): E {
    switch (this.status) {
      case "ok":
        throw new Error("unwrap");
      case "err":
        return this.errValue!;
    }
  }

  map<U>(func: (_: T) => U): Result<U, E> {
    switch (this.status) {
      case "ok":
        return Result.ok(func(this.okValue!));
      case "err":
        return Result.err(this.errValue!);
    }
  }

  mapErr<U>(func: (_: E) => U): Result<T, U> {
    switch (this.status) {
      case "ok":
        return Result.ok(this.okValue!);
      case "err":
        return Result.err(func(this.errValue!));
    }
  }

  andThen<U>(func: (_: T) => Result<U, E>): Result<U, E> {
    switch (this.status) {
      case "ok":
        return func(this.okValue!);
      case "err":
        return Result.err(this.errValue!);
    }
  }

  orElse<U>(func: (_: E) => Result<U, E>): Result<U, E> {
    switch (this.status) {
      case "ok":
        return Result.ok(this.okValue!);
      case "err":
        return func(this.errValue!);
    }
  }
}

export interface LinesAround {
  lines: string[];
  startLineNum: number;
  focusedLineIdx: number;
  col: number;
}

export function getLinesAround(source: string, at: number): Result<LinesAround, undefined> {
  let pos = 0;
  let lineIdx = 0;
  let focusedLineIdx;
  let col;
  let allLines = [];

  while (true) {
    const re = /\r?\n/;
    const nextNewline = source.slice(pos).match(re);

    if (!nextNewline) break;
    const line = source.slice(pos, pos + nextNewline.index);
    allLines.push(line);

    const start = pos;
    const end = start + line.length;

    if (at >= start && at < end) {
      focusedLineIdx = lineIdx;
      col = at - start;
    }

    pos += line.length + nextNewline.length;
    lineIdx += 1;
  }

  if (focusedLineIdx === undefined || col === undefined) return Result.err(undefined);

  const before = Math.max(0, focusedLineIdx - 1);
  const after = Math.min(allLines.length - 1, focusedLineIdx + 1);

  return Result.ok({
    lines: allLines.slice(before, after + 1),
    focusedLineIdx: focusedLineIdx - before,
    startLineNum: before + 1,
    col,
  });
}
